#! /usr/bin/env python
# -*- coding: utf-8 -*-
####################
# Indigo Tapo plugin
# Copyright (c) 2023 Alan Carter. All rights reserved.
#

################################################################################
# Imports
################################################################################

import sys
import os
import json
import time
import datetime
import re
from copy import copy
import ipaddress
from PyP100 import PyP110
from PyP100 import PyL530


################################################################################
# Globals
################################################################################

def updateVar(name, value, folder=0):

    if name not in indigo.variables:
        indigo.variable.create(name, value=value, folder=folder)
    else:
        indigo.variable.updateValue(name, value)


################################################################################
class Plugin(indigo.PluginBase):

    ########################################
    # Class properties
    ########################################

    def __init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs):
     
        indigo.PluginBase.__init__(self, pluginId, pluginDisplayName, pluginVersion, pluginPrefs)
        self.userName = pluginPrefs.get("userName", "")
        self.userPass = pluginPrefs.get("userPass", "")
        self.costPerKwhString = pluginPrefs.get("costPerKwh", "")
        self.showEnergyInNotes = pluginPrefs.get("energyNotes", False)
        self.logComms = pluginPrefs.get("logComms", False)
    
    ########################################
    def __del__(self):
    
        indigo.PluginBase.__del__(self)
        
    ########################################
    def startup(self):
    
        indigo.server.log("Starting TP-Link Tapo plugin")
        if self.costPerKwhString == "":
            self.costPerKwh = 0
        else:
            self.costPerKwh = float(self.costPerKwhString)


    def deviceStartComm(self, dev):
    
        dev.stateListOrDisplayStateIdChanged()
        return


    def shutdown(self):
    
        indigo.server.log("Stopping TP-Link Tapo plugin")
        pass
  
    ########################################
    
    def runConcurrentThread(self):
    
        indigo.server.log("Starting TP-Link Tapo monitoring thread")
        try:
            while True:
                self.getReadings()
                self.sleep(30)
        except self.StopThread:
            self.shutdown()


    ########################################
    # Called when user prefs are changed
    
    def closedPrefsConfigUi(self, valuesDict, userCancelled):
    
        # If the user saved preferences, update logging parameters
        if userCancelled == False:
            costPerKwh = self.costPerKwh
            self.costPerKwhString = valuesDict.get("costPerKwh", "")
            if self.costPerKwhString == "":
                self.costPerKwh = 0
            else:
                self.costPerKwh = float(self.costPerKwhString)
            if self.costPerKwh != costPerKwh:
                indigo.server.log("Electricity cost changed to %sp" % str(self.costPerKwh))

            energyNotes = self.showEnergyInNotes
            self.showEnergyInNotes = valuesDict.get("energyNotes", False)
            if self.showEnergyInNotes != energyNotes:
                if self.showEnergyInNotes:
                    indigo.server.log("Energy will be shown in Notes column")
                else:
                    indigo.server.log("Energy will not be shown in Notes column")

            logComms = self.logComms
            self.logComms = valuesDict.get("logComms", False)
            if self.logComms != logComms:
                if self.logComms:
                    indigo.server.log("Comms logging is on")
                else:
                    indigo.server.log("Comms logging is off")


        
    ########################################
    # Tapo specific functions
    ########################################

    def getReadings(self):
    
        for dev in indigo.devices.iter("self"):
            # Need a try statement here because device does not reply
            #   while being accessed via the Tapo app
#             try:
            tapoDevice = PyP110.P110(dev.address, self.userName, self.userPass)
            response = tapoDevice.getDeviceInfo()
            if self.logComms:
                indigo.server.log("Device info for %s" % dev.name)
                indigo.server.log(str(response))
            onState = response["device_on"]
            if onState:
                dev.updateStateOnServer(key="onOffState", value=True, clearErrorState=True)
                dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
            else:
                dev.updateStateOnServer(key="onOffState", value=False, clearErrorState=True)
                dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
            dev.updateStateOnServer(key="firmware", value=response["fw_ver"])
    
            if dev.deviceTypeId == "tapoBulb":
                if onState:
                    bLevel = int(response["brightness"])
                    wattsB = round(bLevel * 8.7 / 100)
                else:
                    bLevel = 0
                    wattsB = 0
                dev.updateStateOnServer(key="brightnessLevel", value=bLevel, uiValue=str(bLevel)+"%")
            response = tapoDevice.getDeviceUsage()
            if self.logComms:
                indigo.server.log("Device usage for %s" % dev.name)
                indigo.server.log(str(response))
            mT3 =  response["time_usage"]["today"]
            mW3 =  response["time_usage"]["past7"]
            mM3 =  response["time_usage"]["past30"]
            dev.updateStateOnServer(key="minsToday", value=mT3, uiValue=str(mT3)+" mins") 
            dev.updateStateOnServer(key="minsPast7Days", value=mW3, uiValue=str(mW3)+" mins")
            dev.updateStateOnServer(key="minsPast30Days", value=mM3, uiValue=str(mM3)+" mins")
            dev.updateStateOnServer(key="timeToday", value=self.mins2hrs(mT3)) 
            dev.updateStateOnServer(key="timePast7Days", value=self.mins2hrs(mW3))
            dev.updateStateOnServer(key="timePast30Days", value=self.mins2hrs(mM3))
        
            if dev.deviceTypeId == "tapoBulb":
                eT3 =  response["power_usage"]["today"]
                costToday = round(float(eT3) * self.costPerKwh / 100000, 2)
                eW3 =  response["power_usage"]["past7"]
                eM3 =  response["power_usage"]["past30"]
                costPast30Days = round(float(eM3) * self.costPerKwh / 100000, 2) 
                dev.updateStateOnServer(key="energyCurLevel", value=wattsB, uiValue=str(wattsB)+" W")
                dev.updateStateOnServer(key="WhToday", value=eT3, uiValue=str(eT3)+" Wh") 
                dev.updateStateOnServer(key="WhPast7Days", value=eW3, uiValue=str(eW3)+" Wh")
                dev.updateStateOnServer(key="WhPast30Days", value=eM3, uiValue=str(eM3)+" Wh")
                dev.updateStateOnServer(key="costToday", value=costToday, uiValue='£{0:.2f}'.format(costToday))
                dev.updateStateOnServer(key="costPast30Days", value=costPast30Days, uiValue='£{0:.2f}'.format(costPast30Days))
                                            
            dev.setErrorStateOnServer(None)
            self.sleep(2)

            if dev.deviceTypeId == "tapoPlug":
                response = tapoDevice.getEnergyUsage()
                if self.logComms:
                    indigo.server.log("Energy usage for %s" % dev.name)
                    indigo.server.log(str(response))
                wattsString = str(response["current_power"])
                watts = round(float(wattsString)/1000, 1)
                kwhd3 = int(str(response["today_energy"]))
                kwhd = round(kwhd3/1000, 1)
                costToday = round(float(kwhd) * self.costPerKwh / 100, 2)
                kwhm3 = int(str(response["month_energy"]))
                kwhm = round(kwhm3/1000, 1)
                costMonth = round(float(kwhm) * self.costPerKwh / 100, 2)
                dev.updateStateOnServer(key="energyCurLevel", value=watts, uiValue=str(watts)+" W")
                dev.updateStateOnServer(key="kWhDay", value=kwhd, uiValue=str(kwhd)+" kWh")
                dev.updateStateOnServer(key="kWhMonth", value=kwhm, uiValue=str(watts)+" kWh")
                dev.updateStateOnServer(key="costToday", value=costToday, uiValue='£{0:.2f}'.format(costToday))
                dev.updateStateOnServer(key="costMonth", value=costMonth, uiValue='£{0:.2f}'.format(costMonth))
                if self.showEnergyInNotes:
                    dev.description = str(watts)+" W"
                    dev.replaceOnServer()
                elif dev.description != "":
                    dev.description = ""
                    dev.replaceOnServer()
#             except:
#                 dev.setErrorStateOnServer("Not Ready")
#                 if self.logComms:
#                     indigo.server.log("%s comms failed" % dev.name, isError=True)


    def mins2hrs(self, mins):

        hours = mins // 60
        minutes = mins % 60
        return "{}h {}m".format(hours, minutes)
        


    ########################################################
    # Validation methods for plugin prefs and device config
    ########################################################
    
    def validateDeviceConfigUi(self, valuesDict, typeId, devId):    

        errorMsgDict = indigo.Dict()                
        try:
            ip = ipaddress.ip_address(valuesDict["address"])
            return (True, valuesDict)
        except ValueError:
            errorMsgDict["address"] = "Invalid IP Address"                            
            return (False, valuesDict, errorMsgDict)
        


    ########################################
    # Smart plug/bulb action callback
    ########################################
    
    # Main switch action bottleneck called by Indigo Server.
    def actionControlDevice(self, action, dev):
        tapoControl = PyP110.P110(dev.address, self.userName, self.userPass)
        if action.deviceAction == indigo.kDeviceAction.TurnOn:
            try:
                response = tapoControl.turnOn()
                dev.updateStateOnServer("onOffState", True)
                dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOn)
            except:
                indigo.server.log("%s 'on' command failed" % dev.name)
                
        elif action.deviceAction == indigo.kDeviceAction.TurnOff:
            try:
                response = tapoControl.turnOff()
                dev.updateStateOnServer("onOffState", False)
                dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
            except:
                indigo.server.log("%s 'off' command failed" % dev.name)
                
        elif action.deviceAction == indigo.kDeviceAction.SetBrightness and dev.deviceTypeId == "tapoBulb":
            new_brightness = action.actionValue
            if new_brightness >0:
                try:
                    tapoControl = PyL530.L530(dev.address, self.userName, self.userPass)
                    response = tapoControl.setBrightness(new_brightness)
                    dev.updateStateOnServer(key="brightnessLevel", value=new_brightness, uiValue=str(new_brightness)+"%")
                except:
                    indigo.server.log("%s brightness change command failed" % dev.name, isError=True)
            else:
                # Tapo will not accept brightness=0 so turn off lamp instead
                try:
                    response = tapoControl.turnOff()
                    dev.updateStateOnServer("onOffState", False)
                    dev.updateStateImageOnServer(indigo.kStateImageSel.PowerOff)
                except:
                    indigo.server.log("%s 'off' command failed" % dev.name)                
        else:
            indigo.server.log("%s: action not supported" % dev.name, isError=True)


